"""Pool updater updates pools and tokens."""
import logging

from Arbie import PoolValueError
from Arbie.Actions import Action
from Arbie.async_helpers import async_map, run_async
from Arbie.Contracts import BalancerPool, ContractFactory, UniswapPair
from Arbie.Variables import PoolType

logger = logging.getLogger()


class PoolUpdater(Action):
    """Pool Updater updates pools and tokens.

    [Settings]
    input:
        web3: web3
        pools: all_pools
    output:
        new_pools: all_pools
    """

    def __init__(self, config=None):
        self.pair_factory = None
        self.pool_factory = None
        super().__init__(config)

    async def on_next(self, data):
        web3 = data.web3()

        self.pair_factory = ContractFactory(web3, UniswapPair)
        self.pool_factory = ContractFactory(web3, BalancerPool)

        pools = await self._update_pools(data.pools())

        data.new_pools(list(filter(None, pools)))

    def _get_contract(self, address, pool_type):
        if pool_type == PoolType.uniswap:
            return self.pair_factory.load_contract(address=address)
        if pool_type == PoolType.balancer:
            return self.pool_factory.load_contract(address=address)
        raise ValueError("Cannot update pool with unknown type")

    async def _update_pool(self, pool):
        pool_contract = await run_async(
            self._get_contract, pool.address, pool.pool_type
        )
        balances = await pool_contract.get_balances()
        balances_numb = [balance.to_number() for balance in balances]
        try:
            pool.update_balances(balances_numb)
        except PoolValueError as e:
            logger.warn(e)
            return None
        logger.info(f"Updated Pool balance {pool.address}")
        return pool

    async def _update_pools(self, pools):
        logger.info(f"Updating pools, total {len(pools)}")
        return await async_map(self._update_pool, pools)



"""Test balancer contracts."""
import pytest

from Arbie import IERC20TokenError
from Arbie.Contracts import ContractFactory
from Arbie.Contracts.balancer import BalancerFactory, BalancerPool
from Arbie.Contracts.tokens import BadERC20Token, GenericToken
from Arbie.Variables import BigNumber, PoolType

bg10 = BigNumber(10)
bg5 = BigNumber(5)


@pytest.fixture
def pool_factory(deploy_address, w3) -> BalancerFactory:
    return ContractFactory(w3, BalancerFactory).deploy_contract(deploy_address)


@pytest.mark.asyncio
async def test_create_new_pool(pool_factory):
    pool_factory.new_pool()
    assert len(await pool_factory.all_pools()) == 1


@pytest.mark.asyncio
async def test_get_number_of_tokens(pool_factory):
    pool_factory.new_pool()
    pool_factory.new_pool()
    pools = await pool_factory.all_pools()
    assert len(pools) == 2
    assert pools[0].get_number_of_tokens() == 0


def test_bind_weight(pool_factory):
    pool = pool_factory.new_pool()
    with pytest.raises(ValueError):
        amount = BigNumber(5)
        weight = 0.2
        pool.bind("", amount, weight)


@pytest.fixture
async def pool_with_tokens(
    pool_factory: BalancerFactory, dai: GenericToken, weth: GenericToken
) -> BalancerPool:
    weight = 5
    return await pool_factory.setup_pool([dai, weth], [weight, weight], [bg5, bg10])


@pytest.fixture
async def factory_with_bad_token(
    pool_factory: BalancerFactory, dai: GenericToken, bad: BadERC20Token
):
    weight = 5
    await pool_factory.setup_pool(
        [bad, dai], [weight, weight], [bg5, bg10], approve_owner=False
    )
    return pool_factory


@pytest.mark.asyncio
async def test_bind_token_to_pool(pool_with_tokens: BalancerPool):
    tokens = await pool_with_tokens.get_tokens()
    assert len(tokens) == 2
    balances = await pool_with_tokens.get_balances()
    assert balances == [bg5, bg10]
    assert await pool_with_tokens.get_weights() == [0.5, 0.5]


@pytest.mark.asyncio
async def test_create_pool(
    pool_with_tokens: BalancerPool, dai: GenericToken, weth: GenericToken
):
    pool = await pool_with_tokens.create_pool()
    weth_token = await weth.create_token()
    dai_token = await dai.create_token()
    assert pool.pool_type == PoolType.balancer
    assert pool.spot_price(weth_token, dai_token) == 2
    assert pool.balances[0].value == 5
    assert pool.balances[1].value == 10


@pytest.mark.asyncio
async def test_create_bad_pool(factory_with_bad_token: BalancerFactory):
    pools = await factory_with_bad_token.all_pools()
    with pytest.raises(IERC20TokenError):
        await pools[0].create_pool()



"""Test arbie smart contracts."""

import pytest

from Arbie.Actions.arbitrage import ArbitrageFinder
from Arbie.Contracts import GenericToken, UniswapV2Router
from Arbie.Variables import BigNumber


class TestRouter(object):
    @pytest.mark.asyncio
    def test_out_given_in(self, router, trade):
        trade.amount_in = 1
        amount_out = router.check_out_given_in(trade)
        assert amount_out > 1

        profit = ArbitrageFinder(trade).calculate_profit(1)
        assert pytest.approx(1, 1e-5) == amount_out - profit

    @pytest.mark.asyncio
    async def test_swap(
        self, trade, router: UniswapV2Router, weth: GenericToken, dummy_account
    ):
        trade.amount_in = 1
        weth.transfer(dummy_account.address, BigNumber(2))
        weth.set_account(dummy_account)
        balance_before = await weth.balance_of(dummy_account.address)

        assert balance_before > 1

        weth.approve(router.get_address(), BigNumber(2))

        router.set_account(dummy_account)
        assert router.swap(trade)
        amount_out = router.check_out_given_in(trade)

        balance_after = await weth.balance_of(dummy_account.address)
        assert (
            balance_after.to_number() - balance_before.to_number()
            > amount_out - trade.amount_in - 0.001
        )



"""Actions are used for building complex interactions with smart contracts."""

from Arbie.Actions.action import Action, Store  # noqa: F401
from Arbie.Actions.action_tree import ActionTree  # noqa: F401
from Arbie.Actions.arbitrage import Arbitrage  # noqa:F401
from Arbie.Actions.path_finder import PathFinder  # noqa: F401
from Arbie.Actions.pool_finder import PoolFinder  # noqa: F401
from Arbie.Actions.pool_updater import PoolUpdater  # noqa: F401
from Arbie.Actions.redis_state import RedisState  # noqa: F401
from Arbie.Actions.trader import LogTrader, SetUpTrader, Trader  # noqa: F401
from Arbie.Actions.whitelist import Whitelist  # noqa: F401



"""Main application."""

import logging

import Arbie
from Arbie.settings_parser import SettingsParser

logger = logging.getLogger()


class App(object):
    """App is used for configuring and running Arbie."""

    def __init__(self, config):
        logger.info(f"Arbie version {Arbie.__version__}")  # noqa: WPS609
        sp = SettingsParser(config)
        self.store = sp.setup_store()
        self.action_tree = sp.action_tree(self.store)

    async def run(self):
        if self.action_tree is None:
            logger.warning("No actions given in configuration")
            return
        try:
            await self.action_tree.run()
        except Exception as e:
            logger.fatal(e, exc_info=True)
            raise e

    def stop(self):
        self.action_tree.stop()



"""Utility functions for createting large numbers."""


class BigNumber(object):
    """BigNumber is used for creating etheruem friendly numbers."""

    def __init__(self, value, exp=18):
        self.value = int(round(value * 10 ** exp))
        self.exp = exp

    def __eq__(self, other):
        if isinstance(other, BigNumber):
            return self.value == other.value and self.exp == other.exp
        return self.value == other

    def __gt__(self, other):
        if isinstance(other, BigNumber):
            return self.value > other.value
        return self.value > other

    def __truediv__(self, other):
        return self.to_number() / other.to_number()

    @classmethod
    def from_value(cls, value, exp=18):
        bg = cls(0, exp)
        bg.value = int(value)
        return bg

    def to_number(self) -> float:
        return self.value / (10 ** self.exp)



"""Base class for actions."""

from typing import Dict

import yaml

from Arbie import StateError
from Arbie.prometheus import get_prometheus

ADD_TIME = get_prometheus().summary(
    "store_add_time", "Time spent adding items from store"
)
GET_TIME = get_prometheus().summary(
    "store_get_time", "Time spent getting items from store"
)


class Argument(object):
    def __init__(self, default_value):
        self.name = None
        self.value = None
        if default_value == "None":
            return
        if isinstance(default_value, str):
            self.name = default_value
        else:
            self.value = default_value

    def __eq__(self, other):
        return self.name == other.name and self.value == other.value


def parse_settings(settings: Dict):
    if settings is None:
        return None
    parsed_settings = {}
    for key, argument in settings.items():
        parsed_settings[key] = Argument(argument)
    return parsed_settings


def to_none_safe_dict(items) -> Dict[str, str]:
    if items:
        return items
    return {}


class Action(object):
    """Action is a base class for data processing actions.

    Actions are combined to ActionTrees. Actions are configured
    by parsing their __doc__ and comparing with the settings that
    have been given when starting Arbie.

    Everything bellow the settings section is parsed as yaml.
    There are two lists that need to be configured. The input
    list and the output list. These are then mapped to the items
    that can be read to the store.
    [Settings]
    input:
        key_name: 1  # default_value
    output:
        key_name: variable_name
    """

    input_key = "input"
    output_key = "output"

    def __init__(self, config=None):
        self.settings = self._create_settings()
        self.settings = self._update_settings_with_config(config)

    def get_input_settings(self) -> Dict[str, str]:
        return to_none_safe_dict(self.settings[self.input_key])

    def get_output_settings(self) -> Dict[str, str]:
        return to_none_safe_dict(self.settings[self.output_key])

    async def on_next(self, data):
        raise NotImplementedError("Action does not have a on_next statement")

    def _create_settings(self):
        settings = yaml.safe_load(self.__doc__.split("[Settings]\n")[1])

        return {
            self.input_key: parse_settings(settings[self.input_key]),
            self.output_key: parse_settings(settings[self.output_key]),
        }

    def _update_settings_with_config(self, config):
        if config is None:
            return self.settings

        if self.input_key in config and config[self.input_key] is not None:
            self._emplace_settings(config[self.input_key], self.get_input_settings())

        if self.output_key in config and config[self.output_key] is not None:
            self._emplace_settings(config[self.output_key], self.get_output_settings())

        return self.settings

    def _emplace_settings(self, config, settings):
        for key, name in config.items():
            if key not in settings:
                raise ValueError(
                    f"""Argument: {key} is not a valid input/output for action {type(self).__name__},
                        look over your configuration"""
                )
            settings[key] = Argument(name)


def get_value_lambda(value):
    return lambda _: value


class Store(object):
    """Store the state of the Action Tree.

    Making it possible for Actions to share state between each other.
    It also makes it possible for actions to export state.
    """

    def __init__(self, state=None):
        if state is None:
            self.state = {}
        else:
            self.state = state

    def __getitem__(self, key):
        return self.state[key]

    @ADD_TIME.time()
    def add(self, key, item):
        if hasattr(item, "__len__"):  # noqa: WPS421
            get_prometheus().gauge(
                f"{key.lower()}_number", f"Number of {key}"
            ).set(  # noqa: WPS221
                len(item)
            )  # noqa: WPS221
        self.state[key] = item

    @GET_TIME.time()
    def get(self, key):
        return self.state[key]

    def delete(self, key):
        try:
            self.state.delete(key)
        except AttributeError:
            raise StateError("RedisState is required to remove keys from store.")

    def subscribe(self, event_channel):
        try:
            return self.state.subscribe(event_channel)
        except AttributeError:
            raise StateError("RedisState is required to subscribe to store.")

    def publish(self, event_channel, message):
        try:
            self.state.publish(event_channel, message)
        except AttributeError:
            raise StateError("RedisState is required to publish event.")

    def create_input(self, action):
        return self._create_data(
            action.get_input_settings(), action.get_output_settings()
        )

    def _create_data(self, input_settings, output_settings):
        methods = {}
        for key, argument in input_settings.items():
            if argument.name in self.state:
                methods[key] = self._get_lambda(argument.name)
            elif argument.value is not None:
                methods[key] = get_value_lambda(argument.value)
            else:
                raise ValueError(
                    f"Argument {key}, with name {argument.name} not found in state and no default value"
                )

        for key_out, argument_out in output_settings.items():
            methods[key_out] = self._add_lambda(argument_out.name)
        return type("ActionData", (), methods)()

    def _get_lambda(self, key):
        return lambda _: self.get(key)

    def _add_lambda(self, key):
        return lambda _, value: self.add(key, value)



"""settings parser can be used for parsing input yaml."""

import json
import logging

from eth_account import Account
from requests import Session
from requests.adapters import HTTPAdapter
from web3 import Web3, middleware
from web3.gas_strategies.time_based import fast_gas_price_strategy

from Arbie.Actions import ActionTree, RedisState, Store
from Arbie.Contracts import (
    BalancerFactory,
    ContractFactory,
    UniswapFactory,
    UniswapV2Router,
    Weth,
)
from Arbie.Contracts.contract import Network


def to_network(string: str) -> Network:
    if string.lower() == "mainnet":
        return Network.mainnet
    if string.lower() == "kovan":
        return Network.kovan
    return Network.ropsten


class Keys(object):
    actions = "actions"
    address = "address"
    network = "network"
    store = "store"
    type_key = "type"
    variables = "variables"
    value = "value"
    version = "version"
    web3 = "web3"
    action_tree = "action_tree"
    event = "event"
    redis = "redis"
    account = "account"
    key = "key"
    path = "path"


def setup_gas_strategy(w3):
    w3.eth.setGasPriceStrategy(fast_gas_price_strategy)

    w3.middleware_onion.add(middleware.time_based_cache_middleware)
    w3.middleware_onion.add(middleware.latest_block_based_cache_middleware)
    w3.middleware_onion.add(middleware.simple_cache_middleware)


class VariableParser(object):
    """VariableParser parses settings config and adds variable to store."""

    def __init__(self, config, w3_config, account_config):
        self.config = config
        self.w3 = self.set_up_web3(w3_config)
        if account_config is None:
            self.account = None
        else:
            self.account = self._set_up_account(account_config)

    def add_variables(self, store):
        for name, variable_config in self.config.items():
            store.add(name, self._create_variable(variable_config))
        store.add(Keys.web3, self.w3)
        if self.account:
            store.add(Keys.account, self.account)

    def set_up_web3(self, config):
        address = config[Keys.address]

        adapter = HTTPAdapter(
            pool_connections=20, pool_maxsize=20, max_retries=10  # noqa: WPS432
        )  # noqa: WPS432
        session = Session()
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        w3 = Web3(Web3.HTTPProvider(address, session=session))
        setup_gas_strategy(w3)
        if not w3.isConnected():
            raise ConnectionError("Web3 is not connected")

        logging.getLogger().info(f"Connected to Node {address}")
        logging.getLogger().info(f"Current block is {w3.eth.blockNumber}")
        return w3

    def set_up_uniswap(self, config):
        return self._set_up_contracts(config, UniswapFactory)

    def set_up_balancer(self, config):
        return self._set_up_contracts(config, BalancerFactory)

    def set_up_token(self, config):
        return self._set_up_contracts(config, Weth)

    def set_up_router(self, config):
        return self._set_up_contracts(config, UniswapV2Router)

    def _create_variable(self, variable_config):  # noqa: WPS321
        variable_type = variable_config[Keys.type_key]
        if variable_type == "UniswapFactory":
            return self.set_up_uniswap(variable_config)
        if variable_type == "BalancerFactory":
            return self.set_up_balancer(variable_config)
        if variable_type == "Weth":
            return self.set_up_token(variable_config)
        if variable_type == "UniswapV2Router":
            return self.set_up_router(variable_config)
        if variable_type == "float":
            return float(variable_config[Keys.value])
        if variable_type == "int":
            return int(variable_config[Keys.value])
        if variable_type == "str":
            return str(variable_config[Keys.value])
        raise TypeError(f"No rule for creating variable if type {variable_type}")

    def _set_up_contracts(self, config, contract, *kwargs):
        factory = ContractFactory(self.w3, contract)
        if Keys.network in config:
            return factory.load_contract(
                network=to_network(config[Keys.network]), account=self.account
            )
        return factory.load_contract(address=config[Keys.address], account=self.account)

    def _set_up_account(self, account_config):
        with open(account_config[Keys.path], "r") as config_file:
            config = json.load(config_file)
            return Account.from_key(config[Keys.key])


class SettingsParser(object):
    """Settings Parser is used to configure Arbie."""

    def __init__(self, config):
        self.config = config

    def setup_store(self):
        store = self.store()
        if Keys.variables in self.config:
            self._add_variables(store)
        return store

    def store(self):
        if Keys.store in self.config:
            store_config = self.config[Keys.store]
            redis_state = RedisState(store_config[Keys.address])
            return Store(redis_state)
        return Store()

    def action_tree(self, store):
        if Keys.action_tree in self.config:
            return self._setup_action_tree(store, self.config[Keys.action_tree])

    def _setup_action_tree(self, store, config):
        tree = ActionTree.create(config[Keys.actions], store)
        if Keys.event not in config:
            return tree

        events = config[Keys.event]
        if not isinstance(events, list):
            events = [events]
        for event in events:
            tree.register_event(event)
        return tree

    def _add_variables(self, store):
        account_conf = None
        if Keys.account in self.config:
            account_conf = self.config[Keys.account]

        variable_parser = VariableParser(
            self.config[Keys.variables], self.config[Keys.web3], account_conf
        )
        variable_parser.add_variables(store)



"""Test uniswap contracts."""
import pytest

from Arbie.Contracts import ContractFactory
from Arbie.Contracts.tokens import GenericToken, Weth
from Arbie.Variables import BigNumber

bg10 = BigNumber(10)


@pytest.fixture
def token_factory(w3) -> ContractFactory:
    return ContractFactory(w3, GenericToken)


@pytest.fixture
def dai(deploy_address, token_factory) -> GenericToken:
    return token_factory.deploy_contract(deploy_address, "Dai", "DAI", bg10.value)


@pytest.mark.asyncio
async def test_decimals(dai: GenericToken):
    assert await dai.decimals() == 18


class TestToken(object):
    def test_equals(self, dai, token_factory):
        dai2 = token_factory.load_contract(
            owner_address=dai.owner_address, address=dai.get_address()
        )
        assert dai == dai2

    def test_approve(self, dai: GenericToken, deploy_address):
        assert dai.approve(deploy_address, bg10)

    @pytest.mark.asyncio
    async def test_transfer(self, dai: GenericToken, deploy_address, dummy_address):
        dai.approve(deploy_address, bg10)
        dai.transfer(dummy_address, bg10)
        bg = await dai.balance_of(dummy_address)
        assert bg.to_number() == 10

    @pytest.mark.asyncio
    async def test_name(self, dai: GenericToken):
        name = await dai.get_name()
        assert name == "Dai"


class TestWeth(object):
    @pytest.mark.asyncio
    async def test_deposit_withdraw(self, real_weth: Weth, dummy_address):
        real_weth.set_owner_address(dummy_address)
        real_weth.deposit(2)
        assert await real_weth.balance_of(dummy_address) == BigNumber(2)

        real_weth.withdraw(2)
        assert await real_weth.balance_of(dummy_address) == BigNumber(0)



"""Event filter is a helpter class for filtering events with asyncio."""
import logging

from prometheus_async.aio import time

from Arbie.async_helpers import async_map, run_async
from Arbie.prometheus import get_prometheus

GET_ENTRIES = get_prometheus().summary(
    "event_filter_get_entries", "Time for getting entries"
)


class EventTransform(object):
    def __init__(self, event_filter, event_transform):
        self.event_filter = event_filter
        self.event_transform = event_transform

    def run(self):
        events = self.event_filter.get_all_entries()
        return self.event_transform(events)


class EventFilter(object):
    def __init__(self, event, event_transform, from_block, to_block, steps):
        self.from_block = from_block
        self.to_block = to_block
        self.steps = steps
        self.event = event
        self.event_transform = event_transform

    async def find_events(self):
        nmb_event_chunks = int((self.to_block - self.from_block) / self.steps) + 1
        results = await async_map(self._get_entries, range(nmb_event_chunks))

        # results is a list of lists, that we need to flatten
        # https://stackoverflow.com/questions/952914/how-to-make-a-flat-list-out-of-list-of-lists
        return [item for sublist in results for item in sublist]

    async def _get_entries(self, index):
        from_block = index * self.steps + self.from_block
        to_block = from_block + self.steps - 1
        if to_block > self.to_block:
            to_block = self.to_block
        logging.getLogger().info(
            f"Searching for Pools in block range [{from_block}:{to_block}]"
        )
        return await self._get_entries_range(from_block, to_block)

    @time(GET_ENTRIES)
    async def _get_entries_range(self, from_block, to_block):
        event_filter = self.event.createFilter(
            fromBlock=int(from_block), toBlock=int(to_block)
        )
        ev = EventTransform(event_filter, self.event_transform)
        return await run_async(ev.run)



"""System tests for PoolUpdater."""

import pytest

from Arbie.Actions import ActionTree, LogTrader, SetUpTrader, Store, Trader
from Arbie.Contracts import GenericToken, Weth
from Arbie.Variables import BigNumber


def send_eth(web3, from_address, to_address, value):
    tx_hash = web3.eth.sendTransaction(
        {"to": to_address, "from": from_address, "value": BigNumber(value).value}
    )
    return web3.eth.waitForTransactionReceipt(tx_hash, 180)  # noqa: WPS432


@pytest.fixture
def trader_account(w3, weth: GenericToken, router, dummy_account):
    router.set_account(dummy_account)
    weth.transfer(dummy_account.address, BigNumber(2))
    weth.set_account(dummy_account)
    weth.approve(router.get_address(), BigNumber(2))
    return dummy_account


min_profit = "min_profit"

conf_dict = {"input": {min_profit: min_profit}, "output": {}}


@pytest.fixture
def trade_store(w3, router, trade, weth, trader_account):
    store = Store()
    store.add("router", router)
    store.add("filtered_trades", [trade])
    store.add("weth", weth)
    store.add("web3", w3)
    store.add("account", trader_account)
    return store


class TestTrader(object):
    @pytest.mark.asyncio
    async def test_on_next(self, trade_store, trade):
        trade.amount_in = 1
        trade_store.add(min_profit, 0)
        tree = ActionTree(trade_store)
        tree.add_action(Trader(conf_dict))
        await tree.run()
        assert trade_store.get("profit") > 0.036  # noqa: WPS432

    @pytest.mark.asyncio
    async def test_no_profit(self, trade_store, trade):
        trade.amount_in = 1
        tree = ActionTree(trade_store)
        tree.add_action(Trader())
        with pytest.raises(Exception):
            await tree.run()


class TestLogTrader(object):
    @pytest.mark.asyncio
    async def test_on_next(self, trade_store, trade):
        trade.amount_in = 1
        trade_store.add(min_profit, 0)
        tree = ActionTree(trade_store)
        tree.add_action(LogTrader(conf_dict))
        await tree.run()
        assert trade_store.get("profit") > 0.036  # noqa: WPS432


class TestSetUpTrader(object):
    @pytest.fixture
    def trade_store(self, w3, router, real_weth, dummy_account):
        real_weth.set_account(dummy_account)
        store = Store()
        store.add("weth", real_weth)
        store.add("web3", w3)
        store.add("account", dummy_account)
        store.add("router", router)
        return store

    @pytest.mark.asyncio
    async def test_setup_trader(self, trade_store, dummy_account, real_weth: Weth):
        tree = ActionTree(trade_store)
        tree.add_action(SetUpTrader())
        await tree.run()
        assert trade_store.get("balance_weth") == 10
        real_weth.withdraw(10)



"""A store object that uses redis for backend."""

import pickle  # noqa: S403

import redis


def init_redis(address):
    host_and_port = address.split(":")
    if len(host_and_port) != 2:
        raise ValueError(f"Invalid Address to redis server: {address}")
    return redis.Redis(host=host_and_port[0], port=host_and_port[1], db=0)


def _is_collection(key):
    parts = key.split(".")
    return len(parts) == 3 and parts[2].endswith("s")


def _is_item(key):
    parts = key.split(".")
    if len(parts) == 3 and not _is_collection(key):
        return True
    return len(parts) == 4


class RedisState(object):
    """A bridge to access the state stored in redis.

    This makes it possible for different Arbie instances to share state.

    Keys for redis is stored in the following format:
        namespace.version.category.identifier

    Namespace is the name of the Arbie instance running or the name
    of the Arbie instance that we wan to get data from.

    Version is used so that we can change the API with out worrying about breaking
    Arbie instances that rely on old data. When we update how we store data,
    we should update the version. This makes more sence later once we have started
    using protobuff.

    Category contains information regarding what keys are available. If we store
    a list of Tokens then the category would contain the names of the tokens. So
    to get the token we first get the category to figure out which tokens are
    available. A category containing a list of items ends with an s. A categaroy
    is of redis type set.

    Identifier contains the identifier of the actuall data. In case of a Token
    this would be the address of that token.
    """

    def __init__(self, host):
        self.r = init_redis(host)
        self.r.ping()
        self.local_state = {}

    def __getitem__(self, key):
        if _is_collection(key):
            return self._get_collection(key)
        elif _is_item(key):
            return self._get_item(key)
        return self.local_state[key]

    def __setitem__(self, key, value):
        if _is_collection(key):
            self._add_collection(key, value)
        elif _is_item(key):
            self._add_item(key, value)
        else:
            self.local_state[key] = value
            return
        self.publish(key, "updated")

    def __contains__(self, item) -> bool:
        if item is None:
            return False
        if item in self.local_state:
            return True
        return self.r.exists(item) == 1

    def subscribe(self, event_channel):
        p = self.r.pubsub()
        p.psubscribe(event_channel)
        return p

    def publish(self, event_channel, message):
        self.r.publish(event_channel, message)

    def keys(self):
        return self.local_state.keys()

    def delete(self, key):
        pipe = self.r.pipeline()
        if _is_collection(key):
            collection = self.r.smembers(key)
            for item in collection:
                pipe.delete(f"{key}.{item}")
        pipe.delete(key)
        pipe.execute()

    def _get_collection(self, key):
        self._exist_or_raise(key)
        pipe = self.r.pipeline()
        for item in self.r.smembers(key):
            item_name = item.decode("utf-8")
            item_key = f"{key}.{item_name}"
            pipe.get(item_key)
        raw_items = pipe.execute()
        return list(map(lambda i: pickle.loads(i), raw_items))  # noqa: S301

    def _get(self, key):
        self._exist_or_raise(key)
        return self.r.get(key)

    def _get_item(self, key):
        return pickle.loads(self._get(key))  # noqa: S301

    def _add_collection(self, collection_key, collection):
        pipe = self.r.pipeline()
        for item in collection:
            item_key = f"{collection_key}.{item}"
            pipe.set(item_key, pickle.dumps(item))
            pipe.sadd(collection_key, str(item))
        pipe.execute()

    def _add_item(self, key, value):
        self.r.set(key, pickle.dumps(value))

    def _exist_or_raise(self, key):
        if not self.__contains__(key):
            raise KeyError(f"key: {key} was not found in Redis")



"""Help module for web3 tests."""
from typing import List

import pytest

from Arbie.address import dummy_token_generator
from Arbie.Variables import Pool, PoolType, Token

small = 10000
medium = 1000000
large = 100000000


@pytest.fixture
def eth() -> Token:
    return dummy_token_generator("eth", 1)


@pytest.fixture
def dai() -> Token:
    return dummy_token_generator("dai", 300)


@pytest.fixture
def btc() -> Token:
    return dummy_token_generator("btc", 3.0 / 100)


@pytest.fixture
def yam() -> Token:
    return dummy_token_generator("yam", 3000)


@pytest.fixture
def pools(dai, eth, btc, yam) -> List[Pool]:
    return [
        Pool(
            [eth, dai, yam],
            [small / 303.0, small / 0.9, small / 0.1],
            [1 / 3.0, 1 / 3.0, 1 / 3.0],
            PoolType.uniswap,
            0.004,
        ),
        Pool(
            [eth, btc],
            [large / 305.0, large / 10000],
            [5 / 6, 1 / 6],
            PoolType.uniswap,
            0.01,
        ),
        Pool(
            [eth, dai, btc],
            [medium / 301.0, medium / 1.1, medium / 10020],
            [1 / 2.0, 1 / 4.0, 1 / 4.0],
            PoolType.uniswap,
            0.004,
        ),
        Pool(
            [dai, yam],
            [small / 1.1, small / 0.1],
            [1 / 2.0, 1 / 2.0],
            PoolType.balancer,
            0.001,
        ),
    ]



"""Unittest of ActionTree."""

import pytest
import yaml

from Arbie.Actions import Action, ActionTree, Store


class DummyAction(Action):
    """
    Dummy description for dummy action.

    [Settings]
    input:
    output:
    """


store = Store()


class TestActionTree(object):
    def test_create(self):
        tree = ActionTree.create({"PathFinder": {}}, store)
        assert len(tree.actions) == 1

    def test_create_bad_arg(self):
        config = """
        PathFinder:
            input:
                non_existing: []
        """
        with pytest.raises(ValueError):
            ActionTree.create(yaml.safe_load(config), store)

    def test_create_bad_action(self):
        with pytest.raises(ValueError):
            ActionTree.create({"NonExistingAction": {}}, store)

    def test_create_dummy_action(self):
        config = """
        PathFinder:
        DummyAction:
        """

        tree = ActionTree.create(
            yaml.safe_load(config), store, [("DummyAction", DummyAction)]
        )
        assert len(tree.actions) == 2



"""Arbie.

Arbie is configured through a Yaml file.

Usage:
  Arbie -f config.yml [-l arbie.log.txt]
  Arbie (-h | --help)
  Arbie (-v | --version)

Options:
  -f --file=config.yml  Load configuration file.

  -l --log=arbie.log    Path to log files [default: arbie.log]

  -h --help             Show this screen.
  -v --version          Show version.

"""
import asyncio
import logging
from logging import handlers

import yaml
from docopt import docopt

import Arbie
from Arbie.app import App

max_log_size = int(10e6)  # noqa: WPS432


def setup_logging(log_file: str, severity=logging.INFO):
    formater = "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
    logging.basicConfig(
        level=severity, format=formater, datefmt="%m-%d %H:%M"
    )  # noqa: WPS323

    root_logger = logging.getLogger()

    file_handler = handlers.RotatingFileHandler(
        log_file, maxBytes=max_log_size, backupCount=5
    )
    file_handler.setFormatter(logging.Formatter(formater))
    file_handler.setLevel(severity)
    root_logger.addHandler(file_handler)


def main(argv=None):
    arguments = docopt(__doc__, argv, version=Arbie.__version__)  # noqa: WPS609

    logging.getLogger().info(f"arguments: {arguments}")

    setup_logging(str(arguments["--log"]))

    config_path = str(arguments["--file"])
    config = None
    with open(config_path, "r") as config_file:
        config = yaml.safe_load(config_file)

    app = App(config)

    try:
        asyncio.run(app.run())
    except Exception as ex:
        logging.getLogger().error(ex)
        raise ex



"""Path finder contains Actions for finding paths between nodes."""

import logging
from typing import List

import networkx as nx

from Arbie.Actions import Action
from Arbie.Variables import Pool, Token, Tokens, Trade
from Arbie.Variables.graph import FilteredTradingGraph, TradingGraph


class Node(object):
    """A edge in a cycle."""

    def __init__(self, token: Token, pool: Pool = None):
        self.token = token
        self.pool = pool

    def __eq__(self, other):
        return self.token == other.token


class Cycle(object):
    """A cycle of nodes with the ratio of the cycle."""

    def __init__(self, nodes: List[Node], ratio):
        self.nodes = nodes
        self.ratio = ratio


class CycleFinder(object):
    """Algorithm for finding cycles and the ratio of the cycle.

    Recursive Depth first search is done over all nodes. When reaching the start
    node a cycle has been found. While searching also remembers the weights for
    taking a cycle. This is useful for sorting the cycles after trade value.
    """

    def __init__(self, graph: nx.DiGraph, start_node: Token, max_depth):
        self.max_depth = max_depth
        self.graph = graph
        self.start_node = Node(start_node)

    def find_all_cycles(self):
        cycles = []
        self._visit_neighbours(cycles, [self.start_node], self.start_node, 1)
        return cycles

    def _visit_neighbours(
        self,
        found_cycles: List[Cycle],
        visited: List[Node],
        current_node: Node,
        ratio_to_current_node,
    ):

        for _, next_token, data in self.graph.edges(current_node.token, data=True):
            ratio_to_next_node = ratio_to_current_node * data["weight"]
            next_node = Node(next_token, data["object"])
            self._visit_node(found_cycles, visited, next_node, ratio_to_next_node)

    def _visit_node(
        self,
        found_cycles: List[Cycle],
        visited: Tokens,
        current_node: Node,
        ratio_to_current_node,
    ):
        if current_node == self.start_node:
            # We have come back to start. Append cycle to result
            found_cycles.append(Cycle(visited + [current_node], ratio_to_current_node))
            logging.getLogger(f"Found cycle {len(found_cycles)}!")
            return

        if current_node in visited:
            # We have found a cycle back to the current node. Stop
            return

        if len(visited) > self.max_depth:
            # We have gone to deap lets stop.
            return

        self._visit_neighbours(
            found_cycles, visited + [current_node], current_node, ratio_to_current_node
        )


def create_trade(cycle: Cycle) -> Trade:
    path = [cycle.nodes[0].token]
    pools = []
    for node in cycle.nodes[1:]:
        pools.append(node.pool)
        path.append(node.token)
    return Trade(pools, path, ratio=cycle.ratio)


class PathFinder(Action):
    """Find all trades from a trading graph.

    Also find the ratio of taking that trade.

    [Settings]
    input:
        pools: pools
        weth: None
        min_liquidity: 1
        max_depth: 5
    output:
        cycles: all_cycles
        trades: all_trades
    """

    async def on_next(self, data):
        logging.getLogger().info("Searching for path")
        graph = FilteredTradingGraph(TradingGraph(data.pools()), data.min_liquidity())
        finder = CycleFinder(
            graph.graph, await data.weth().create_token(1), data.max_depth()
        )
        cycles = sorted(finder.find_all_cycles(), key=lambda x: x.ratio)
        logging.getLogger().info(f"Found {len(cycles)} cycles")
        data.cycles(cycles)
        trades = []
        for cycle in cycles:
            trades.append(create_trade(cycle))
        data.trades(trades)



"""Utility functions for interacting with Arbie.sol."""

import logging

from Arbie.Contracts.contract import Contract
from Arbie.Contracts.tokens import GenericToken
from Arbie.Variables import BigNumber, Trade

logger = logging.getLogger()


class UniswapV2Router(Contract):
    name = "UniswapV2Router02"
    protocol = "uniswap"

    def approve(self, weth: GenericToken):
        if weth.allowance(self.get_address()) < BigNumber(10e6):  # noqa: WPS432
            return weth.approve(self.get_address(), BigNumber(10e8))  # noqa: WPS432
        return True

    def check_out_given_in(self, trade: Trade):
        path_address = list(map(lambda t: t.address, trade.path))
        amount_out = self.contract.functions.getAmountsOut(
            BigNumber(trade.amount_in).value, path_address
        ).call()
        return BigNumber.from_value(amount_out[-1]).to_number()

    def swap(self, trade):
        gas = self._estimate_gas_swap(trade)
        transaction = self._swap_transaction(trade)
        return self._transact_status(transaction, gas=gas)

    def _swap_transaction(self, trade):
        path = list(map(lambda t: t.address, trade.path))
        return self.contract.functions.swapExactTokensForTokens(
            BigNumber(trade.amount_in).value,
            BigNumber(trade.amount_in).value,
            path,
            self._get_account(),
            # Require trades to be executed in 120 seconds
            self.w3.eth.getBlock("latest").timestamp + 120,  # noqa: WPS432
        )

    def _estimate_gas_swap(self, trade):
        transaction = self._swap_transaction(trade)
        # Lets use a bit more gas then required, since
        # some tokens like $VLO actually suck gas
        # out of the transaction
        return int(self._estimate_gas(transaction) * 1.05) + 1  # noqa: WPS432



"""Trades contain variables for trading."""

from Arbie.Variables.pool import Pools
from Arbie.Variables.token import Balance, Tokens


class Trade(object):
    def __init__(self, pools: Pools, path: Tokens, ratio=None, amount_in=None):
        self.pools = pools
        self.path = path
        self.amount_in = amount_in
        self.profit = None
        self.balance = None
        self.ratio = ratio

    def __len__(self):
        return len(self.path)

    def __str__(self):
        return f"""
Trade(
    pools:{self.pools},
    path:{self.path},
    amount_in:{self.amount_in},
    profit:{self.profit},
    balance:{self.balance},
    ratio:{self.ratio})"""  # noqa: WPS221

    def __iter__(self):
        return iter(self._generator())

    def __getitem__(self, i):
        return self.pools[i], self.path[i], self.path[i + 1]  # noqa: WPS221

    def set_balance(self, balance: Balance):
        self.balance = balance

    def _generator(self):
        yield from (self[i] for i, _ in enumerate(self.pools))



"""Unittest of arbitrage."""

import pytest

from Arbie.Actions import ActionTree, Store
from Arbie.Actions.arbitrage import Arbitrage, ArbitrageFinder
from Arbie.address import dummy_token_generator
from Arbie.Variables import Pool, Trade


@pytest.fixture
def tokens(dai, eth):
    return [dai, eth]


@pytest.fixture
def trade1(tokens, dai, eth):
    pool1 = Pool(tokens, [400, 1], [0.5, 0.5])
    pool2 = Pool(tokens, [410, 1], [0.5, 0.5])
    return Trade([pool1, pool2], [dai, eth, dai])


@pytest.fixture
def trade2(tokens, dai, eth):
    pool1 = Pool(tokens, [400, 1], [0.9, 0.1])
    pool2 = Pool(tokens, [410, 1], [0.1, 0.9])
    return Trade([pool1, pool2], [dai, eth, dai])


class TestArbitrage(object):
    """Test Arbitrage."""

    def test_find_arbitrage(self, trade1):
        amount, _ = ArbitrageFinder(trade1).find_arbitrage()
        assert amount == pytest.approx(2.48456731316587)  # noqa: WPS432

    def test_find_arbitrage_unbalanced(self, trade2):
        amount, _ = ArbitrageFinder(trade2).find_arbitrage()
        assert amount == pytest.approx(27.8547574719045)  # noqa: WPS432

    def test_find_arbitrage_no_opportunity(self, tokens, dai, eth):
        pool1 = Pool(tokens, [400, 1], [0.9, 0.1])
        pool2 = Pool(tokens, [410, 1], [0.1, 0.9])
        trade = [Trade(pool1, eth, dai), Trade(pool2, dai, eth)]

        with pytest.raises(ValueError):
            ArbitrageFinder(trade).find_arbitrage()

    def test_calc_optimal_arbitrage_no_opportunity(self, tokens, dai, eth):
        pool1 = Pool(tokens, [400, 1], [0.9, 0.1])
        pool2 = Pool(tokens, [410, 1], [0.1, 0.9])
        trade = Trade([pool1, pool2], [eth, dai, eth])

        with pytest.raises(ValueError) as e:
            ArbitrageFinder(trade).calculate_optimal_arbitrage()
            assert e.message == "No arbitrage opportunity found."

    def test_find_arbitrage_wrong_token(self, tokens, dai):
        pool1 = Pool(tokens, [400, 1], [0.9, 0.1])
        pool2 = Pool(tokens, [410, 1], [0.1, 0.9])
        sai = dummy_token_generator("sai", 300.0)
        trade = Trade([pool1, pool2], [dai, sai, dai])

        with pytest.raises(ValueError):
            ArbitrageFinder(trade).find_arbitrage()


class TestArbitrageAction(object):
    @pytest.mark.asyncio
    async def test_on_next(self, trade1, trade2):
        store = Store()
        store.add("all_trades", [trade1, trade2])
        tree = ActionTree(store)
        tree.add_action(Arbitrage())
        await tree.run()

        assert len(store.get("filtered_trades")) == 2



"""Unittest for pool updater."""

from unittest.mock import MagicMock

import pytest

from Arbie.Actions import ActionTree, PoolUpdater, Store

pytestmark = pytest.mark.asyncio


class TestPoolUpdater(object):
    async def test_on_next(self, pools):
        store = Store()
        store.add("web3", MagicMock())
        store.add("all_pools", pools)
        tree = ActionTree(store)
        tree.add_action(PoolUpdater())
        with pytest.raises(ValueError):
            await tree.run()



"""Trader contains actions for executing trades."""

import logging

from Arbie import TransactionError
from Arbie.Actions import Action
from Arbie.Variables import BigNumber, PoolType

logger = logging.getLogger()


class BalanceChecker(object):
    def __init__(self, web3, weth):
        self.web3 = web3
        self.weth = weth

    async def check(self, address):
        amount_weth = await self.weth.balance_of(address)
        amount_eth = BigNumber.from_value(self.web3.eth.getBalance(address))
        return amount_eth.to_number(), amount_weth.to_number()

    async def check_total(self, address):
        return sum(await self.check(address))

    async def check_and_convert(self, trader_address, min_eth, min_weth, max_weth):
        amount_eth, amount_weth = await self.check(trader_address)

        # Already have sufficient amount setup
        if amount_eth > min_eth and amount_weth > min_weth:
            return amount_eth, amount_weth

        # We don't have sufficient supply
        if amount_eth + amount_weth < min_eth + min_weth:
            raise ValueError(
                f"Not enough liquidity eth: {amount_eth}, min_eth: {min_eth}, weth {amount_weth}, min_weth: {min_weth}"
            )

        # ETH is good but not weth
        if amount_eth > min_eth + min_weth:
            self._to_weth(amount_eth, min_eth, max_weth)
        else:
            self._to_eth(amount_eth, amount_weth, min_eth, min_weth)

        return await self.check(trader_address)

    def _to_weth(self, amount_eth, min_eth, max_weth):
        max_deposit = amount_eth - min_eth
        max_deposit = min(max_deposit, max_weth)
        status = self.weth.deposit(max_deposit)
        if not status:
            raise TransactionError("Failed to deposit eth for weth")

    def _to_eth(self, amount_eth, amount_weth, min_eth, min_weth):
        max_withdraw = amount_weth - min_weth
        max_withdraw = min(max_withdraw, min_eth - amount_eth)
        status = self.weth.withdraw(max_withdraw)
        if not status:
            raise TransactionError("Failed to withdraw weth for eth")


class SetUpTrader(Action):
    """SetUp Trader account for trading.

    [Settings]
    input:
        web3: web3
        weth: weth
        router: router
        min_eth: 1
        min_weth: 2
        max_weth: 10
        account: account
    output:
        balance_eth: balance_eth
        balance_weth: balance_weth
    """

    async def on_next(self, data):
        balance_checker = BalanceChecker(data.web3(), data.weth())
        trader_account = data.account()
        amount_eth, amount_weth = await balance_checker.check_and_convert(
            trader_account.address,
            data.min_eth(),
            data.min_weth(),
            data.max_weth(),
        )
        router = data.router()
        if not router.approve(data.weth()):
            raise Exception("Failed to authorize arbie to spend tokens.")

        logger.info(
            f"Finished setting up trader, eth: {amount_eth}, weth: {amount_weth}"
        )

        data.balance_eth(amount_eth)
        data.balance_weth(amount_weth)


def _is_trade_uni(trade):
    for pool in trade.pools:
        if pool.pool_type is not PoolType.uniswap:
            return False
    return True


def _perform_trade(trade, router, min_profit):
    if not _is_trade_uni(trade):
        return False

    amount_out = router.check_out_given_in(trade)
    profit = amount_out - trade.amount_in
    logger.info(
        f"Checking trade with profit {profit}, amount_in: {trade.amount_in}, amount out: {amount_out}"
    )
    if profit > min_profit:
        logger.info(f"Executing trade {trade}")
        return router.swap(trade)
    return False


def perform_trade(data, amount_weth):
    router = data.router()
    min_profit = data.min_profit()
    for trade in data.trades():
        # Make sure that we don't try to trade with more weth then we have
        trade.amount_in = min(trade.amount_in, amount_weth)
        if _perform_trade(trade, router, min_profit):
            return True
    return False


class Trader(Action):
    """Find optimal arbitrage opportunity for a list sorted trades.

    Remove all trades that are not profitable.

    [Settings]
    input:
        web3: web3
        router: router
        trades: filtered_trades
        min_profit: 0.3
        weth: weth
        account: account
    output:
        profit: profit
    """

    async def on_next(self, data):  # noqa: WPS210
        trader_account = data.account()
        balance_checker = BalanceChecker(data.web3(), data.weth())
        amount_eth, amount_weth = await balance_checker.check(trader_account.address)
        balance_pre = amount_weth + amount_eth
        logger.info(f"amount_eth: {amount_eth}, amount_weth: {amount_weth}")

        if not perform_trade(data, amount_weth):
            raise Exception("No trade performed")

        balance_post = await balance_checker.check_total(trader_account.address)

        data.profit(balance_post - balance_pre)
        logger.info("Finished trading")


class LogTrader(Action):
    """Log arbitrage opportunity for a list sorted trades.

    Remove all trades that are not profitable.

    [Settings]
    input:
        web3: web3
        router: router
        trades: filtered_trades
        min_profit: 0.3
        weth: weth
    output:
        profit_per_trade: profit_per_trade
        profit: profit
    """

    async def on_next(self, data):  # noqa: WPS210
        value = self.log_trade(data)
        profit = sum(value)
        profit_per_trade = profit / len(value)

        logger.info(f"Total profit: {profit}, profit per trade: {profit_per_trade}")

        data.profit_per_trade(profit_per_trade)
        data.profit(profit)

        logger.info("Finished trading")

    def log_trade(self, data):
        router = data.router()
        min_profit = data.min_profit()
        profits = []
        for trade in data.trades():
            profit = self._log_trade(trade, router, min_profit)
            if profit:
                profits.append(profit)
        return profits

    def _log_trade(self, trade, router, min_profit):
        amount_out = router.check_out_given_in(trade)
        profit = amount_out - trade.amount_in
        if profit > min_profit:
            logger.info(
                f"Executing trade with profit {profit}, amount_in: {trade.amount_in}, amount out: {amount_out}"
            )
            return profit



"""Module for gettings tokens from Coingecko."""

import logging
from urllib.parse import urljoin

import requests

from Arbie.async_helpers import async_map, run_async

logger = logging.getLogger()

COINGECKO_URL = "https://api.coingecko.com"

COINS_URL = urljoin(COINGECKO_URL, "api/v3/coins/list")


class Coingecko(object):
    def __init__(self, batch_size=5, timeout=0.6):
        self.batch_size = batch_size
        self.timeout = timeout

    async def coins(self):
        ids = await self.ids()
        if ids:
            return await self.coins_from_ids(ids)
        return []

    async def coins_from_ids(self, ids):
        response = await self._coin_urls(ids)
        addresses = set(map(self._parse_eth_coin, response))
        addresses.remove(None)
        return list(addresses)

    async def ids(self):
        rs = await self._get(COINS_URL)
        if rs.ok:
            return list(map(lambda i: i["id"], rs.json()))

    async def _coin_urls(self, ids):
        urls = list(map(self._coin_url, ids))
        return await async_map(self._get, urls, self.batch_size, self.timeout)

    async def _coin_ticker(self, coin_id):
        url = self._coin_url(coin_id)
        rs = await self._get(url)
        return self._parse_eth_coin(rs)

    def _coin_url(self, coin_id):
        return urljoin(COINGECKO_URL, f"api/v3/coins/{coin_id}/tickers")

    def _parse_eth_coin(self, response):
        if not response.ok:
            return None
        tickers = self._tickers(response)
        if (
            tickers is None
            or tickers["target"] != "ETH"
            or not self._ok(tickers["is_anomaly"])
        ):
            return None
        address = tickers["base"].lower()
        eth_address_length = 42  # noqa: WPS432
        if len(address) == eth_address_length:
            return address

    def _tickers(self, response):
        tickers = response.json()["tickers"]
        if tickers:
            return tickers[0]

    def _ok(self, is_anomaly):
        return not is_anomaly

    async def _get(self, url):
        logger.info(f"Requesting endpoing {url}")
        return await run_async(requests.get, url)



"""Common configuration for Action system tests."""

import pytest

from Arbie.Actions import RedisState, Store


@pytest.fixture
def redis_state(redis_server):
    return RedisState(redis_server)


@pytest.fixture
def redis_store(redis_state):
    return Store(redis_state)



"""Whitelist contains actions for finding good tokens."""

from Arbie.Actions import Action
from Arbie.Services import Coingecko


class Whitelist(Action):
    """SetUp Trader account for trading.

       Coingecko is used for finding whitelisted tokens it has a limit
       of 10 requests/second. However in reality it is much lower.
       The base settings allow for 1 requests every 0.6 seconds
       is equal to 1.67 req/sec. In total we need to do ~6000 request
       which will take ~60 min.
    [Settings]
    input:
        requests: 1
        delay: 0.6
    output:
        whitelist: whitelist
    """

    async def on_next(self, data):
        data.whitelist(await Coingecko(1, 0.6).coins())  # noqa: WPS432



"""Unittest for the action structure."""
from Arbie.Variables.graph import FilteredTradingGraph, TradingGraph


def test_create(pools):
    graph = TradingGraph(pools)
    assert len(graph) == 4
    assert len(graph.get_edges()) == 16


def test_create_filter(pools):
    f_graph = FilteredTradingGraph(TradingGraph(pools), 0)
    assert len(f_graph) == 4
    assert len(f_graph.get_edges()) == 10



"""Unittest trader."""

import asyncio
from unittest.mock import MagicMock

import pytest

from Arbie import TransactionError
from Arbie.Actions.trader import BalanceChecker
from Arbie.Variables import BigNumber


def amount_to_future(amount):
    future = asyncio.Future()
    future.set_result(BigNumber(amount))
    return future


def setup_eth(amounts):
    mock = MagicMock()
    mock.eth = mock
    mock.getBalance.side_effect = list(
        map(lambda a: BigNumber(a).value, amounts)
    )  # noqa: WPS221
    return mock


def setup_weth(amounts):
    mock = MagicMock()
    mock.balance_of.side_effect = list(map(amount_to_future, amounts))
    return mock


def setup_mocks(eth_amounts, weth_amounts):
    return setup_eth(eth_amounts), setup_weth(weth_amounts)


some_address = "any address"


class TestBalanceChecker(object):
    @pytest.mark.asyncio
    async def test_check(self):
        checker = BalanceChecker(*setup_mocks([1], [2]))
        eth, weth = await checker.check(some_address)
        assert eth == 1
        assert weth == 2

    @pytest.mark.asyncio
    async def test_low_eth(self):
        checker = BalanceChecker(*setup_mocks([1], [0]))
        with pytest.raises(ValueError):
            await checker.check_and_convert(some_address, 1, 2, 10)

    @pytest.mark.asyncio
    async def test_high_eth_and_weth(self):
        checker = BalanceChecker(*setup_mocks([12], [11]))
        eth, weth = await checker.check_and_convert(some_address, 1, 2, 10)
        assert eth == 12
        assert weth == 11

    @pytest.mark.asyncio
    async def test_high_eth(self):
        eth_mock, weth_mock = setup_mocks([12, 2], [0, 10])
        checker = BalanceChecker(eth_mock, weth_mock)

        eth, weth = await checker.check_and_convert(some_address, 1, 2, 10)
        assert eth == 2
        assert weth == 10

        weth_mock.deposit.assert_called_once_with(10)

    @pytest.mark.asyncio
    async def test_high_weth(self):
        eth_mock, weth_mock = setup_mocks([0.5, 1], [10, 9.5])
        checker = BalanceChecker(eth_mock, weth_mock)

        eth, weth = await checker.check_and_convert(some_address, 1, 2, 10)
        assert eth == 1
        assert weth == 9.5

        weth_mock.withdraw.assert_called_once_with(0.5)

    @pytest.mark.asyncio
    async def test_high_eth_transaction_error(self):
        eth_mock, weth_mock = setup_mocks([12], [0])
        weth_mock.deposit.return_value = False

        checker = BalanceChecker(eth_mock, weth_mock)
        with pytest.raises(TransactionError):
            await checker.check_and_convert(some_address, 1, 2, 10)

    @pytest.mark.asyncio
    async def test_high_weth_transaction_error(self):
        eth_mock, weth_mock = setup_mocks([0.5], [10])
        weth_mock.withdraw.return_value = False

        checker = BalanceChecker(eth_mock, weth_mock)
        with pytest.raises(TransactionError):
            await checker.check_and_convert(some_address, 1, 2, 10)



"""Help module for web3 tests."""
import asyncio
import json

import pytest
from eth_account import Account
from web3 import Web3

from Arbie.async_helpers import async_map
from Arbie.Contracts import (
    BalancerFactory,
    ContractFactory,
    UniswapFactory,
    UniswapV2Router,
    Weth,
)
from Arbie.Contracts.tokens import BadERC20Token, GenericToken
from Arbie.settings_parser import setup_gas_strategy
from Arbie.Variables import BigNumber, Trade


@pytest.fixture
def web3_server(request):
    return request.config.getoption("--web3_server")


@pytest.fixture
def redis_server(request):
    return request.config.getoption("--redis_server")


@pytest.fixture
def w3(web3_server):
    return Web3(Web3.HTTPProvider(web3_server))


@pytest.fixture
def w3_with_gas_strategy(web3_server):
    w3 = Web3(Web3.HTTPProvider(web3_server))
    setup_gas_strategy(w3)
    return w3


@pytest.fixture
def deploy_address(w3) -> str:
    return w3.eth.accounts[0]


@pytest.fixture
def dummy_address(w3) -> str:
    return w3.eth.accounts[1]


@pytest.fixture
def empty_account(w3) -> Account:
    with open("Brig/Trader/test_account.json", "r") as config_file:
        config = json.load(config_file)
        return Account.from_key(config["key"])


@pytest.fixture()
def dummy_account(w3, empty_account) -> Account:
    balance = w3.eth.getBalance(empty_account.address)
    required = 11 - BigNumber.from_value(balance).to_number()
    required = max(0, required)

    tx_recip = w3.eth.sendTransaction(
        {
            "from": w3.eth.accounts[2],
            "to": empty_account.address,
            "value": BigNumber(required).value,
        }
    )
    w3.eth.waitForTransactionReceipt(tx_recip)
    return empty_account


@pytest.fixture
def token_factory(w3) -> ContractFactory:
    return ContractFactory(w3, GenericToken)


large_number = BigNumber(10e14)


@pytest.fixture
async def dai(deploy_address, token_factory) -> GenericToken:
    token = token_factory.deploy_contract(
        deploy_address, "Dai", "DAI", large_number.value
    )
    await token.approve_owner()
    return token


@pytest.fixture
async def weth(deploy_address, token_factory) -> GenericToken:
    token = token_factory.deploy_contract(
        deploy_address, "Weth", "WETH", large_number.value
    )
    await token.approve_owner()
    return token


@pytest.fixture
async def yam(deploy_address, token_factory) -> GenericToken:
    token = token_factory.deploy_contract(
        deploy_address, "yam", "YAM", large_number.value
    )
    await token.approve_owner()
    return token


@pytest.fixture
async def blocked_token(deploy_address, token_factory) -> GenericToken:
    token = token_factory.deploy_contract(
        deploy_address, "shit", "SHI", large_number.value
    )
    await token.approve_owner()
    return token


@pytest.fixture
async def wbtc(deploy_address, token_factory) -> GenericToken:
    token = token_factory.deploy_contract(
        deploy_address, "Wbtc", "WBTC", large_number.value
    )
    await token.approve_owner()
    return token


@pytest.fixture
def bad(deploy_address, w3) -> BadERC20Token:
    return ContractFactory(w3, BadERC20Token).deploy_contract(
        deploy_address, large_number.value
    )


@pytest.fixture
def real_weth(w3, deploy_address):
    return ContractFactory(w3, Weth).deploy_contract(deploy_address)


small = 10e4
medium = 10e6
large = 10e8


@pytest.fixture
async def pool_factory(
    dai: GenericToken,
    weth: GenericToken,
    yam: GenericToken,
    wbtc: GenericToken,
    bad,
    w3,
    deploy_address,
) -> BalancerFactory:
    factory = ContractFactory(w3, BalancerFactory).deploy_contract(deploy_address)

    f1 = factory.setup_pool(
        [weth, dai, yam],
        [5, 5, 5],
        [
            BigNumber(small / 303.0),
            BigNumber(small / 0.9),
            BigNumber(small / 0.1),
        ],
    )
    f2 = factory.setup_pool(
        [bad, dai],
        [5, 5],
        [
            BigNumber(100),
            BigNumber(small / 0.9),
        ],
        approve_owner=False,
    )
    factory.new_pool()

    f3 = factory.setup_pool(
        [weth, wbtc],
        [5, 1],
        [
            BigNumber(5 * large / 301.0),
            BigNumber(large / 10000),
        ],  # noqa: WPS221
    )
    f4 = factory.setup_pool(
        [weth, dai, wbtc],
        [2, 1, 1],
        [
            BigNumber(2 * medium / 301.0),
            BigNumber(medium / 1.1),
            BigNumber(large / 10020),
        ],
    )

    await asyncio.gather(f1, f2, f3, f4)

    return factory


@pytest.fixture
def factory(deploy_address, w3) -> UniswapFactory:
    return ContractFactory(w3, UniswapFactory).deploy_contract(
        deploy_address, deploy_address
    )


@pytest.fixture
async def pair_factory(  # noqa: WPS210, WPS217
    factory,
    dai: GenericToken,
    weth: GenericToken,
    yam: GenericToken,
    wbtc: GenericToken,
    blocked_token: GenericToken,
    bad,
    w3,
    deploy_address,
) -> UniswapFactory:
    await factory.setup_pair(
        [wbtc, dai],
        [
            BigNumber(medium / 20000.0),
            BigNumber(medium),
        ],
    )
    await factory.setup_pair(
        [dai, yam],
        [
            BigNumber(medium / 1.1),
            BigNumber(medium / 0.1),
        ],  # noqa: WPS221
    )
    await factory.setup_pair(
        [weth, dai],
        [BigNumber(medium / 310), BigNumber(medium)],
    )
    await factory.setup_pair(
        [wbtc, weth],
        [
            BigNumber(medium / 10000),
            BigNumber(medium / 285),
        ],
    )

    await factory.setup_pair(
        [blocked_token, weth],
        [
            BigNumber(medium / 10000),
            BigNumber(medium / 285),
        ],
    )

    await factory.create_pair(weth, bad)
    await factory.create_pair(weth, yam)
    return factory


@pytest.fixture
def whitelist(  # noqa: WPS210, WPS217
    dai: GenericToken, weth: GenericToken, yam: GenericToken, wbtc: GenericToken
):
    return [
        dai.get_address().lower(),
        weth.get_address().lower(),
        yam.get_address().lower(),
        wbtc.get_address().lower(),
    ]


@pytest.fixture
def router(w3_with_gas_strategy, deploy_address, weth, factory):
    return ContractFactory(w3_with_gas_strategy, UniswapV2Router).deploy_contract(
        deploy_address, factory.get_address(), weth.get_address()
    )


async def create_pool(pool):
    return await pool.create_pool()


async def create_token(token):
    return await token.create_token()


@pytest.fixture
async def pairs(factory, weth, dai, wbtc):
    p0 = await factory.setup_pair(
        [weth, dai],
        [
            BigNumber(large / 300.0),
            BigNumber(large),
        ],
    )

    p1 = await factory.setup_pair(
        [wbtc, dai],
        [
            BigNumber(large / 10000.0),
            BigNumber(large),
        ],
    )

    p2 = await factory.setup_pair(
        [wbtc, weth],
        [
            BigNumber(large / 10000.0),
            BigNumber(large / 285.0),
        ],
    )
    return await async_map(create_pool, [p0, p1, p2])


@pytest.fixture
async def path(weth, dai, wbtc):
    raw_path = [weth, dai, wbtc, weth]
    return await async_map(create_token, raw_path)


@pytest.fixture
def trade(pairs, path):
    return Trade(pairs, path)



"""unit test for Coingecko."""

from unittest.mock import MagicMock

import pytest
from pytest_mock.plugin import MockerFixture

from Arbie.Services import Coingecko

pytestmark = pytest.mark.asyncio

coins = [
    {"id": "1ai", "symbol": "1ai", "name": "1AI"},
    {"id": "velo-token", "symbol": "VLO", "name": "VELO Token"},
    {"id": "1inch", "symbol": "1inch", "name": "1inch"},
]

coin_1a1 = {
    "name": "1AI",
    "tickers": [
        {
            "base": "1AI",
            "target": "BTC",
            "market": {
                "name": "Vindax",
                "identifier": "vindax",
                "has_trading_incentive": False,
            },
            "last": 1e-8,
            "volume": 51070,
            "converted_last": {"btc": 1e-8, "eth": 2.919e-7, "usd": 0.00036442},
            "converted_volume": {"btc": 0.0005107, "eth": 0.01490667, "usd": 18.61},
            "trust_score": None,
            "bid_ask_spread_percentage": 50,
            "timestamp": "2021-01-17T02:39:22+00:00",
            "last_traded_at": "2021-01-17T02:39:22+00:00",
            "last_fetch_at": "2021-01-18T02:17:05+00:00",
            "is_anomaly": False,
            "is_stale": True,
            "trade_url": "https://vindax.com/exchange-base.html?symbol=1AI_BTC",
            "token_info_url": None,
            "coin_id": "1ai",
            "target_coin_id": "bitcoin",
        }
    ],
}

coin_vlo = {
    "name": "VELO Token",
    "tickers": [
        {
            "base": "0X98AD9B32DD10F8D8486927D846D4DF8BAF39ABE2",
            "target": "ETH",
            "market": {
                "name": "Uniswap (v2)",
                "identifier": "uniswap",
                "has_trading_incentive": False,
            },
            "last": 0.00000243692314918803,
            "volume": 932296.695673962,
            "converted_last": {"btc": 9.41e-8, "eth": 0.00000244, "usd": 0.00333838},
            "converted_volume": {"btc": 0.08776486, "eth": 2.274565, "usd": 3112.36},
            "trust_score": "green",
            "bid_ask_spread_percentage": 0.614416,
            "timestamp": "2021-01-20T05:32:35+00:00",
            "last_traded_at": "2021-01-20T05:32:35+00:00",
            "last_fetch_at": "2021-01-20T06:01:50+00:00",
            "is_anomaly": False,
            "is_stale": False,
            "trade_url": "https://app.uniswap.org/#/swap?outputCurrency=0x98ad9b32dd10f8d8486927d846d4df8baf39abe2",
            "token_info_url": "https://info.uniswap.org/token/0x98ad9b32dd10f8d8486927d846d4df8baf39abe2",
            "coin_id": "velo-token",
            "target_coin_id": "ethereum",
        }
    ],
}

coin_1inch = {
    "name": "1inch",
    "tickers": [
        {
            "base": "0X111111111117DC0AA78B770FA6A738034120C302",
            "target": "ETH",
            "market": {
                "name": "1inch Liquidity Protocol",
                "identifier": "one_inch_liquidity_protocol",
                "has_trading_incentive": False,
            },
            "last": 0.0013669431402898,
            "volume": 4600000.73782359,
            "converted_last": {"btc": 0.0000527, "eth": 0.00136693, "usd": 1.87},
            "converted_volume": {"btc": 242.4, "eth": 6288, "usd": 8592901},
            "trust_score": "green",
            "bid_ask_spread_percentage": 0.602722,
            "timestamp": "2021-01-20T06:04:08+00:00",
            "last_traded_at": "2021-01-20T06:04:08+00:00",
            "last_fetch_at": "2021-01-20T06:04:08+00:00",
            "is_anomaly": True,  # Modified from original inorder to test anomaly detection
            "is_stale": False,
            "token_info_url": None,
            "coin_id": "1inch",
            "target_coin_id": "ethereum",
        },
    ],
}


def bad_request(endpoint):
    mock = MagicMock()
    mock.ok = True
    return mock


def load_data(endpoint):
    mock = MagicMock()
    mock.ok = True
    json = None
    if "coins/list" in endpoint:
        json = coins
    elif "velo" in endpoint:
        json = coin_vlo
    elif "1inch" in endpoint:
        json = coin_1inch
    elif "1ai" in endpoint:
        json = coin_1a1
    mock.json.return_value = json
    return mock


@pytest.fixture
def request_mock(mocker: MockerFixture):
    mock = mocker.patch("Arbie.Services.coingecko.requests")
    mock.get.side_effect = load_data
    return mock


class TestCoingecko(object):
    async def test_coins(self, request_mock: MagicMock):
        addresses = await Coingecko().coins()
        assert len(addresses) == 1
        assert addresses[0] == "0x98ad9b32dd10f8d8486927d846d4df8baf39abe2"

    async def test_bad_request(self, request_mock: MagicMock):
        request_mock.get.side_effect = bad_request
        addresses = await Coingecko().coins()
        assert not addresses



"""Test uniswap contracts."""
import asyncio

import pytest

from Arbie import IERC20TokenError
from Arbie.Contracts.tokens import BadERC20Token, GenericToken
from Arbie.Contracts.uniswap import UniswapFactory, UniswapPair
from Arbie.Variables import BigNumber, PoolType

bg10 = BigNumber(10)
bg5 = BigNumber(5)

pytestmark = pytest.mark.asyncio


@pytest.fixture
async def factory_with_pair(factory, dai, weth) -> UniswapFactory:
    await factory.create_pair(dai, weth)
    return factory


async def test_get_all_pairs_length(factory):
    assert await factory.all_pairs_length() == 0


async def test_create_pair(factory_with_pair):
    assert await factory_with_pair.all_pairs_length() == 1


async def test_get_all_pairs(factory_with_pair):
    assert len(await factory_with_pair.all_pairs()) == 1


@pytest.fixture
async def pair(factory_with_pair) -> UniswapPair:
    pairs = await factory_with_pair.all_pairs()
    return pairs[0]


async def test_get_weights(pair):
    assert await pair.get_balances() == [0, 0]


async def test_mint(
    pair: UniswapPair, dai: GenericToken, weth: GenericToken, deploy_address
):
    assert dai.transfer(pair.get_address(), bg10)
    assert weth.transfer(pair.get_address(), bg10)
    assert pair.mint(deploy_address)
    assert await pair.get_balances() == [bg10, bg10]


async def test_create_pool(
    pair: UniswapPair, dai: GenericToken, weth: GenericToken, deploy_address
):
    dai.transfer(pair.get_address(), bg5)
    weth.transfer(pair.get_address(), bg10)
    pair.mint(deploy_address)
    pool = await pair.create_pool()
    tokens = await asyncio.gather(weth.create_token(), dai.create_token())

    assert pool.pool_type == PoolType.uniswap
    assert pool.spot_price(tokens[0], tokens[1]) == 2
    balances = pool.get_balances(tokens[0], tokens[1])
    assert balances[0] == 10
    assert balances[1] == 5


async def test_create_bad_pool(
    factory: UniswapFactory, bad: BadERC20Token, dai: GenericToken
):
    pair = await factory.create_pair(bad, dai)
    with pytest.raises(IERC20TokenError):
        await pair.create_pool()



"""Unittest for settings parser."""
from unittest.mock import MagicMock

import pytest
import yaml

from Arbie import StateError
from Arbie.settings_parser import Keys, SettingsParser


@pytest.fixture
def config():
    raw_conf = """
        version: "1.0"
        store:
            address: host_name:port

        web3:
            address: url:port

        variables:
            my_weth_token:
                type: Weth
                address: "0xdac17f958d2ee523a2206206994597c13d831ec7"
            uniswap_factory:
                type: UniswapFactory
                address: "0xdac17f958d2ee523a2206206994597c13d831ec7"
            balancer_factory:
                type: BalancerFactory
                address: "0xdac17f958d2ee523a2206206994597c13d831ec7"
            router:
                type: UniswapV2Router
                address: "0xdac17f958d2ee523a2206206994597c13d831ec7"
            amount:
                type: float
                value: 26.3
            start_block:
                type: int
                value: 1337
            token_name:
                type: str
                value: ethereum

        action_tree:
            event:
                arbie.1.pools
            actions:
                PathFinder:
                    input:
                        weth: my_weth_token
                    output:
                        trades: all_trades
        """
    return yaml.safe_load(raw_conf)


def mock_web3(mocker):
    mocker.patch("Arbie.settings_parser.Web3")


class TestSettingsParser(object):
    def test_redis_store(self, config, mocker):
        mocker.patch("Arbie.settings_parser.RedisState")

        sp = SettingsParser(config)
        store = sp.store()
        assert isinstance(store.state, MagicMock)

    def test_local_store(self, config):
        del config[Keys.store]  # noqa: WPS420

        sp = SettingsParser(config)
        store = sp.store()
        assert isinstance(store.state, dict)

    def test_set_up_store(self, config, mocker):
        mock_web3(mocker)
        del config[Keys.store]  # noqa: WPS420

        sp = SettingsParser(config)
        store = sp.setup_store()
        assert len(store.state) == 8

    def test_set_up_store_no_variables(self, config, mocker):
        mock_web3(mocker)
        del config[Keys.store]  # noqa: WPS420
        del config[Keys.variables]  # noqa: WPS420

        sp = SettingsParser(config)
        store = sp.setup_store()
        assert not store.state

    def test_set_up_action_tree(self, config, mocker):
        mock_web3(mocker)
        mocker.patch("Arbie.settings_parser.RedisState")

        sp = SettingsParser(config)
        store = sp.setup_store()
        tree = sp.action_tree(store)
        assert len(tree.actions) == 1

    def test_subscribe_raises(self, config, mocker):
        mock_web3(mocker)
        del config[Keys.store]  # noqa: WPS420

        sp = SettingsParser(config)
        store = sp.setup_store()
        with pytest.raises(StateError):
            sp.action_tree(store)



"""Test whitelist."""

import pytest
from pytest_mock.plugin import MockerFixture

from Arbie.Actions import ActionTree, Store, Whitelist
from Arbie.Services import Coingecko


class TestWhitelist(object):
    @pytest.mark.asyncio
    async def test_on_next(self, mocker: MockerFixture):
        mocker.patch.object(
            Coingecko,
            "coins",
            return_value=[
                "0xdac17f958d2ee523a2206206994597c13d831ec7",
                "0x514910771af9ca656af840dff83e8264ecf986ca",
            ],
        )
        store = Store()
        tree = ActionTree(store)
        tree.add_action(Whitelist())
        await tree.run()

        assert len(store.get("whitelist")) == 2



"""Helper functions for interacting with eth addresses."""

import os

from Arbie.Variables.token import Token


def dummy_address_generator() -> str:
    return f"0x{os.urandom(20).hex()}"  # noqa: WPS432


def dummy_token_generator(name, value=None) -> Token:
    return Token(name, dummy_address_generator(), value)



"""Main init."""

from Arbie.exception import (  # noqa: F401
    DeployContractError,
    IERC20TokenError,
    PoolValueError,
    StateError,
    TransactionError,
)
from Arbie.settings_parser import SettingsParser  # noqa: F401

__version__ = "0.8.2"  # noqa: WPS410
__version_info__ = tuple(
    int(i) for i in __version__.split(".") if i.isdigit()
)  # noqa: WPS221



"""Systemtests of app."""

import asyncio

import pytest
import yaml
from pytest_mock.plugin import MockerFixture

from Arbie.app import App
from Arbie.Contracts import GenericToken
from Arbie.Services import Coingecko
from Arbie.Variables import BigNumber


class Result(object):
    pool_finder_pools = "PoolFinder.1.pools"
    pool_finder_tokens = "PoolFinder.1.tokens"
    whitelist_addresses = "Whitelist.1.addresses"
    pool_updater_pools = "PoolUpdater.1.pools"
    arbitrage_filtered_trades = "Arbitrage.1.filtered_trades"
    trader_profit = "Trader.1.profit"


pytestmark = pytest.mark.asyncio


async def wait_and_stop(tree, key):
    while True:
        if key in tree.store.state:
            tree.stop()
            return
        await asyncio.sleep(0.1)


async def wait_and_run(app):
    await asyncio.sleep(0.2)
    await app.run()


def replace_base_conf(config, base_config):
    split_conf = config.split("action_tree:")
    return f"""
{base_config}

action_tree:
{split_conf[1]}
            """


def setup_config(path, base_config):
    with open(f"Brig/{path}", "r") as f:
        return replace_base_conf(f.read(), base_config)


class TestApp(object):
    @pytest.fixture
    def base_config(
        self,
        web3_server,
        redis_server,
        weth,
        pool_factory,
        pair_factory,
        router,
        deploy_address,
    ):
        return f"""
store:
    address: {redis_server}
web3:
    address: {web3_server}
variables:
    weth:
        type: Weth
        address: '{weth.get_address()}'
    uniswap_factory:
        type: UniswapFactory
        address: '{pair_factory.get_address()}'
    balancer_factory:
        type: BalancerFactory
        address: '{pool_factory.get_address()}'
    router:
        type: UniswapV2Router
        address: '{router.get_address()}'
        """  # noqa: WPS221

    @pytest.fixture
    def base_config_with_account(self, base_config):
        split_conf = base_config.split("variables:")
        return f"""
{split_conf[0]}
account:
    path: Brig/Trader/test_account.json
variables:
{split_conf[1]}
"""

    @pytest.fixture
    def pool_finder_config(self, base_config):
        conf = setup_config("PoolFinder/pool_finder.yml", base_config)
        split_conf = conf.split("input:")
        return f"""
{split_conf[0]}
      input:
        balancer_start: 0
{split_conf[1]}
"""

    @pytest.fixture
    def path_finder_config(self, base_config):
        return setup_config("PathFinder/path_finder.yml", base_config)

    @pytest.fixture
    def pool_updater_config(self, base_config):
        return setup_config("PoolUpdater/pool_updater.yml", base_config)

    @pytest.fixture
    def trader_config(self, base_config_with_account):
        return setup_config("Trader/trader.yml", base_config_with_account)

    @pytest.fixture
    def pool_finder(self, pool_finder_config, mocker: MockerFixture, whitelist):
        mocker.patch.object(Coingecko, "coins", return_value=whitelist)

        config = yaml.safe_load(pool_finder_config)
        app = App(config)
        assert len(app.action_tree.actions) == 2
        yield app
        app.store.delete(Result.pool_finder_pools)
        app.store.delete(Result.pool_finder_tokens)
        app.store.delete(Result.whitelist_addresses)

    @pytest.fixture
    def pool_updater(self, pool_updater_config):
        config = yaml.safe_load(pool_updater_config)
        app = App(config)
        yield app
        app.store.delete(Result.pool_updater_pools)

    @pytest.fixture
    def path_finder(self, path_finder_config):
        config = yaml.safe_load(path_finder_config)
        app = App(config)
        yield app
        app.store.delete(Result.arbitrage_filtered_trades)

    @pytest.fixture
    def trader(self, trader_config, weth: GenericToken, dummy_account):
        weth.transfer(dummy_account.address, BigNumber(4))
        config = yaml.safe_load(trader_config)
        app = App(config)
        yield app
        app.store.delete(Result.trader_profit)

    @pytest.mark.slow
    async def test_full_pipeline(self, pool_finder, pool_updater, path_finder, trader):
        await asyncio.gather(
            wait_and_run(pool_finder),
            pool_updater.run(),
            wait_and_stop(pool_updater, Result.pool_updater_pools),
            path_finder.run(),
            wait_and_stop(path_finder, Result.arbitrage_filtered_trades),
            trader.run(),
            wait_and_stop(trader, Result.trader_profit),
        )

        assert len(pool_finder.store.get(Result.pool_finder_pools)) == 4
        assert len(pool_finder.store.get(Result.pool_finder_tokens)) == 3
        assert len(pool_finder.store.get(Result.whitelist_addresses)) == 4
        assert len(pool_updater.store.get(Result.pool_updater_pools)) == 4
        assert len(path_finder.store.get(Result.arbitrage_filtered_trades)) == 1
        assert trader.store.get(Result.trader_profit) > 3.278



"""Unittest for Trade."""
import pytest

from Arbie.Variables import Pool, Trade


@pytest.fixture
def tokens(dai, eth):
    return dai, eth


class TestTrade(object):
    @pytest.fixture
    def trade(self, tokens, eth, dai):  # noqa: WPS442
        pool1 = Pool(tokens, [400, 1], [0.5, 0.5])
        pool2 = Pool(tokens, [410, 1], [0.5, 0.5])
        return Trade([pool1, pool2], [dai, eth, dai])

    def test_create(self, trade):
        assert len(trade.pools) == 2
        assert len(trade.path) == 3

    def test_get(self, trade, eth, dai):
        _, token_in, token_out = trade[0]
        assert token_in == dai
        assert token_out == eth

    def test_iter(self, trade, eth, dai):
        pools = []
        for pool, _in, _out in trade:
            pools.append(pool)
        assert len(pools) == 2



"""Unittest of __main__.py."""
import pytest
from pytest_mock.plugin import MockerFixture

from Arbie.__main__ import main
from Arbie.Actions import Store


@pytest.fixture
def config_file() -> str:
    return """
    action_tree:
        actions:
            PathFinder:
                input:
                    weth: weth
                    min_liquidity: 4
                output:
                    cycles: found_cycles
            Arbitrage:
    """


@pytest.fixture
def store() -> Store:
    store = Store()
    store.add("pools", None)
    store.add("weth", None)
    return store


def setup_mocks(mocker, config_file):
    mocker.patch("Arbie.__main__.handlers.RotatingFileHandler")
    mocker.patch("Arbie.__main__.logging.getLogger")
    if config_file is not None:
        mocker.patch("builtins.open", mocker.mock_open(read_data=config_file))


def run_main():
    main(["-f", "giberich.yml"])


class TestMain(object):
    def test_config_not_found(self, mocker: MockerFixture):
        setup_mocks(mocker, None)
        with pytest.raises(FileNotFoundError):
            run_main()

    def test_setup_and_run(self, config_file: str, store: Store, mocker: MockerFixture):
        setup_mocks(mocker, config_file)
        run_mock = mocker.patch("Arbie.settings_parser.ActionTree.run")

        run_main()
        assert run_mock.called

    def test_fail_on_run(self, config_file: str, store: Store, mocker: MockerFixture):
        setup_mocks(mocker, config_file)
        run_mock = mocker.patch("Arbie.settings_parser.ActionTree.run")
        run_mock.side_effect = ValueError("Failed to run action")

        with pytest.raises(ValueError):
            run_main()

    def test_key_not_in_store(self, config_file: str, mocker: MockerFixture):
        setup_mocks(mocker, config_file)

        with pytest.raises(ValueError):
            run_main()



"""Unittest of redis state."""
import pickle  # noqa: S403
from unittest.mock import MagicMock

import pytest
from pytest_mock.plugin import MockerFixture

from Arbie.Actions.redis_state import RedisState

item_key = "pool_finder.1.unit_of_account"
collection_key = "pool_finder.1.pools"
collection_item_key = "pool_finder.1.pools.0xAb12C"


class TestRedisState(object):
    def test_get(self, redis_state, mocker: MockerFixture):
        mock_collection = mocker.patch.object(
            RedisState, "_get_collection", return_value=[]
        )
        mock_item = mocker.patch.object(
            RedisState, "_get_item", return_value=MagicMock()
        )

        with pytest.raises(KeyError):
            redis_state[""]
        assert isinstance(redis_state[collection_key], list)
        assert mock_collection.called
        assert isinstance(redis_state[collection_item_key], MagicMock)
        assert mock_item.called
        assert isinstance(redis_state[item_key], MagicMock)

    def test_add(self, redis_state, mocker: MockerFixture):
        mock_collection = mocker.patch.object(RedisState, "_add_collection")
        mock_item = mocker.patch.object(RedisState, "_add_item")

        redis_state["my_key"] = None
        assert len(redis_state.local_state.keys()) == 1

        redis_state[collection_key] = None
        assert mock_collection.called

        redis_state[item_key] = None
        assert mock_item.called

    def test_get_item_raises(self, redis_state, redis_mock):
        redis_mock.exists.return_value = 0
        with pytest.raises(KeyError):
            redis_state[item_key]
        assert redis_mock.exists.called

    def test_get_collection_raises(self, redis_state, redis_mock):
        redis_mock.exists.return_value = 0
        with pytest.raises(KeyError):
            redis_state[collection_key]
        assert redis_mock.exists.called

    def test_get_collection_item_raises(self, redis_state, redis_mock):
        redis_mock.exists.return_value = 0
        with pytest.raises(KeyError):
            redis_state[collection_item_key]
        assert redis_mock.exists.called

    def test_get_keys(self, redis_state):
        redis_state["item"] = 1
        assert len(redis_state.keys()) == 1

    def test_in_redis_state(self, redis_state, redis_mock):
        redis_mock.exists.side_effect = Exception
        assert None not in redis_state

    def test_get_item(self, redis_state, redis_mock):
        redis_mock.exists.return_value = 1
        redis_mock.get.return_value = pickle.dumps(1337)
        assert redis_state[item_key] == 1337



"""Tests for BigNumber."""

from Arbie.Variables import BigNumber


def test_create_big_number():
    bg = BigNumber(5)
    assert bg.to_number() == 5


def test_equals():
    assert BigNumber(5) == BigNumber(5)


def test_list_equals():
    assert [BigNumber(5)] == [BigNumber(5)]


def test_not_equal():
    assert BigNumber(5) != BigNumber(6)


def test_not_list_equals():
    assert [BigNumber(5)] != [BigNumber(6)]



"""System tests for ActionTree."""
import asyncio

import pytest

from Arbie.Actions import Action, ActionTree, Store


class ActionA(Action):
    """Dummy Action for testing.

    [Settings]
    input:
        times_ran_old: actionA.1.times_ran
    output:
        times_ran: actionA.1.times_ran
        my_result: actionA.1.result
    """

    async def on_next(self, data):
        data.times_ran(data.times_ran_old() + 1)
        data.my_result(1337)


class ActionB(Action):
    """Dummy Action for testing.

    [Settings]
    input:
        times_ran_old: actionB.1.times_ran
        someone_elses_result: actionA.1.result
    output:
        times_ran: actionB.1.times_ran
        my_result: result
    """

    async def on_next(self, data):
        data.times_ran(data.times_ran_old() + 1)
        data.my_result(data.someone_elses_result() + 1)


a_times_ran = "actionA.1.times_ran"
b_times_ran = "actionB.1.times_ran"
a_result = "actionA.1.result"
b_result = "result"

redis_channel = a_result


@pytest.fixture
def tree_a(redis_store: Store):
    tree = ActionTree(redis_store)
    tree.add_action(ActionA())

    redis_store.add(a_times_ran, 0)
    yield tree
    redis_store.delete(a_times_ran)
    redis_store.delete(a_result)


@pytest.fixture
def tree_b(redis_store: Store):
    tree = ActionTree(redis_store)
    tree.add_action(ActionB())

    redis_store.add(b_times_ran, 0)
    yield tree
    redis_store.delete(b_times_ran)


pytestmark = pytest.mark.asyncio


async def wait_and_stop(tree):
    await asyncio.sleep(0.2)
    tree.stop()


class TestActionTree(object):
    async def test_subscribe_no_publisher(self, tree_a, redis_store):
        await tree_a.run()
        await tree_a.run()
        assert redis_store.get(a_times_ran) == 2

    async def test_subscribe(self, redis_store: Store, tree_a):
        tree_a.register_event("my_channel")
        redis_store.publish("my_channel", "new pool added")
        await asyncio.gather(tree_a.run(), wait_and_stop(tree_a))
        assert redis_store.get(a_times_ran) == 1

    # Whenever a result is added to redis
    # a message with that variables name is
    # published to the channel with the
    # same name as the variable.
    async def test_subscribe_publish(
        self, redis_store: Store, tree_a: ActionTree, tree_b: ActionTree
    ):
        tree_b.register_event(redis_channel)

        await asyncio.gather(tree_b.run(), tree_a.run(), wait_and_stop(tree_b))
        assert redis_store.get(b_times_ran) == 1
        assert redis_store.get(a_times_ran) == 1
        assert redis_store.get(a_result) == 1337
        assert redis_store.get(b_result) == 1338



"""arbitrage can be used to find to arbitrage opertunity between two Pools."""

import logging
from typing import Tuple

from sympy import nsolve, symbols

from Arbie.Actions import Action
from Arbie.Variables import Trade

logger = logging.getLogger()
x = symbols("x")


class ArbitrageFinder(object):
    def __init__(self, trade: Trade):
        self.trade = trade

    def find_arbitrage(self) -> Tuple[float, float]:
        if len(self.trade) < 3:
            raise ValueError(
                "Can only find arbitrage opportunity between at leat 3 tokens"
            )

        if not self.token_in_pools():
            raise ValueError("Tokens does not exist in pools")

        trade_input = self.calculate_optimal_arbitrage()
        profit = self.calculate_profit(trade_input)

        return trade_input, profit

    def token_in_pools(self) -> bool:
        for pool, token_in, token_out in self.trade:
            if token_in not in pool.tokens or token_out not in pool.tokens:
                return False
        return True

    def trade_expr(self):
        return self.expr_builder(lambda inner, expr: inner.subs(x, expr))

    def arbitrage_expr(self):
        return self.expr_builder(lambda inner, expr: inner.subs(x, expr) - x)

    def expr_builder(self, subs_expr):  # noqa: WPS210
        i = 0
        expr = None
        for pool, token_in, token_out in self.trade:
            inner_expr = pool.out_given_in_expr(token_in, token_out)
            if i > 0:
                expr = subs_expr(inner_expr, expr)
            else:
                expr = inner_expr
                i = i + 1
        return expr

    def arbitrage_diff_expr(self):
        return self.arbitrage_expr().diff(x)

    def calculate_optimal_arbitrage(self) -> float:
        sol = nsolve(self.arbitrage_diff_expr(), 0)
        if sol <= 0:
            raise AssertionError("No arbitrage opportunity found.")
        return sol

    def calculate_profit(self, value) -> float:
        expr = self.trade_expr()
        return expr.subs(x, value) - value

    def _initial_expr(self):
        pool, trade_in, trade_out = self.trade[0]
        return pool.out_given_in_expr(trade_in, trade_out)


def _update_oportunity(raw_trades, nmb_trades):
    trades = []
    for index, arb in enumerate(raw_trades):
        amount_in = None
        profit = None
        try:
            amount_in, profit = ArbitrageFinder(arb).find_arbitrage()
        except (AssertionError, ValueError, TypeError) as e:
            logger.warning(f"No arbitrage found {e}")
            continue

        arb.amount_in = amount_in
        arb.profit = profit
        trades.append(arb)
        logger.info(
            f"Found opportunity {index}:{len(raw_trades)}"
            + " with profit {arb.profit} for {arb.amount_in} eth"
        )
        if index > nmb_trades:
            break

    return trades


class Arbitrage(Action):
    """Find optimal arbitrage opportunity for a list sorted trades.

    Remove all trades that are not profitable.

    [Settings]
    input:
        trades: all_trades
        nmb_trades: 100
    output:
        out_trades: filtered_trades
    """

    async def on_next(self, data):
        trades = _update_oportunity(data.trades(), data.nmb_trades())

        sorted_trade = sorted(
            trades, key=lambda trade: trade.profit / trade.amount_in, reverse=True
        )
        logger.info(f"Top trade has {sorted_trade[0].profit}")
        data.out_trades(sorted_trade)



"""Contracts enable interacton with ETH contracts."""
from Arbie.Contracts.balancer import BalancerFactory, BalancerPool  # noqa: F401
from Arbie.Contracts.contract import Contract, ContractFactory, Network  # noqa: F401
from Arbie.Contracts.event_filter import EventFilter  # noqa: F401
from Arbie.Contracts.tokens import GenericToken, IERC20Token, Weth  # noqa: F401
from Arbie.Contracts.uniswap import UniswapFactory, UniswapPair  # noqa: F401
from Arbie.Contracts.uniswap_router import UniswapV2Router  # noqa: F401



"""Common fixtures for Actions."""
from unittest.mock import MagicMock

import pytest
from pytest_mock.plugin import MockerFixture

from Arbie.Actions import RedisState


@pytest.fixture
def redis_mock(mocker: MockerFixture) -> MagicMock:
    mock = MagicMock()
    mocker.patch("Arbie.Actions.redis_state.redis.Redis", return_value=mock)
    return mock


@pytest.fixture
def redis_state(redis_mock):
    state = RedisState("good.host.org:1337")
    assert redis_mock.ping.called
    return state



"""Helper functions for working with async."""

import asyncio
from concurrent.futures import ThreadPoolExecutor

thread_pool = ThreadPoolExecutor(max_workers=15)  # noqa: WPS432


def chunked(seq, chunk_size):
    yield from (
        seq[index : index + chunk_size]
        for index in range(0, len(seq), chunk_size)  # noqa: WPS518
    )


async def async_map_chunk(function, seq):
    tasks = [function(i) for i in seq]
    return await asyncio.gather(*tasks)


async def async_map(function, seq, chunk_size=100, wait_time=0):
    res = []
    for chunk in chunked(seq, chunk_size):
        res.extend(await async_map_chunk(function, chunk))
        await asyncio.sleep(wait_time)
    return res


async def run_async(function, *args):
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(thread_pool, function, *args)



"""Unittest of pool."""
import pytest

from Arbie import PoolValueError
from Arbie.address import dummy_address_generator, dummy_token_generator
from Arbie.Variables import Pool


@pytest.fixture
def tokens(dai, eth):
    return [dai, eth]


@pytest.fixture
def pool(tokens):
    return Pool(tokens, [4000, 10], [0.5, 0.5])


class TestPool(object):
    def test_init_multi(self, tokens):
        new_tokens = tokens + [dummy_token_generator("sai", 100)]
        pool = Pool(
            new_tokens, [4e3, 10, 10e5], [1 / 3.0, 1 / 3.0, 1 / 3.0]  # noqa: WPS221
        )  # noqa: WPS221
        assert pool.spot_price(new_tokens[0], new_tokens[1]) == 400
        assert pool.spot_price(new_tokens[2], new_tokens[1]) == 100000

    def test_init_fee_raises(self, tokens):
        with pytest.raises(PoolValueError):
            Pool(tokens, [400, 1], [0.75, 0.25], fee=2)

    def test_init_fee(self, tokens):
        address = dummy_address_generator()
        pool = Pool(tokens, [400, 1], [0.75, 0.25], address=address)
        assert pool.address == address

    def test_pool_bad_weights(self, tokens):
        with pytest.raises(PoolValueError):
            Pool(tokens, [400, 1], [0.7, 0.25])

    def test_spot_price(self, pool, dai, eth):
        assert pool.spot_price(dai, eth) == 400  # noqa: WPS432

    def test_out_give_in(self, pool, dai, eth):
        assert pool.out_given_in(eth, dai, 1) == pytest.approx(
            363.636363636364
        )  # noqa: WPS432

    def test_in_give_out(self, pool, dai, eth):
        assert pool.in_given_out(dai, eth, 1) == pytest.approx(
            444.444444444444
        )  # noqa: WPS432



"""System tests for RedisState."""
import pickle  # noqa: S403

import pytest

from Arbie.Actions.redis_state import RedisState
from Arbie.address import dummy_token_generator

collection_key = "pool_finder.1.pools"
item_key = "pool_finder.1.pools.0xAb12C"


@pytest.fixture
def token():
    return dummy_token_generator("my_token")


@pytest.fixture
def tokens():
    return [dummy_token_generator("token1"), dummy_token_generator("token2")]


class TestRedisState(object):
    @pytest.fixture
    def redis_item(self, redis_state, token):
        redis_state.r.set(item_key, pickle.dumps(token))
        yield None
        redis_state.delete(item_key)

    @pytest.fixture
    def redis_collection(self, redis_state: RedisState, tokens):
        redis_state[collection_key] = tokens
        yield None
        redis_state.delete(collection_key)

    def test_get_empty_state(self, redis_state):
        with pytest.raises(KeyError):
            redis_state[collection_key]
        with pytest.raises(KeyError):
            redis_state[item_key]

    def test_get_item(self, redis_state, redis_item, token):
        assert redis_state[item_key] == token

    def test_get_collection(self, redis_state: RedisState, redis_collection, tokens):
        collection = redis_state[collection_key]

        for t in tokens:
            collection.index(t)

    def test_add_and_get(self, redis_state: RedisState, token):
        redis_state[item_key] = token
        token_round_trip = redis_state[item_key]
        redis_state.delete(item_key)
        assert token_round_trip == token



"""Utility functions for interacting with Tokens."""

import asyncio
import logging

from Arbie.Contracts.contract import Contract
from Arbie.Variables import BigNumber, Token

token_protocol = "tokens"  # noqa: S105


class IERC20Token(Contract):
    name = "IERC20"
    protocol = token_protocol

    def __eq__(self, other):
        return self.get_address() == other.get_address()

    async def decimals(self) -> int:
        return await self._call_async(self.contract.functions.decimals())

    async def balance_of(self, owner: str) -> BigNumber:
        result = await asyncio.gather(
            self._call_async(self.contract.functions.balanceOf(owner)), self.decimals()
        )
        return BigNumber.from_value(result[0], result[1])

    def transfer(self, to: str, bg_number: BigNumber) -> bool:
        transaction = self.contract.functions.transfer(to, bg_number.value)
        return self._transact_status(transaction)

    def allowance(self, to: str):
        return self.contract.functions.allowance(self._get_account(), to).call()

    def approve(self, spender: str, bg_number: BigNumber) -> bool:
        transaction = self.contract.functions.approve(spender, bg_number.value)
        return self._transact_status(transaction)

    async def approve_owner(self):
        bg = await self.balance_of(self.owner_address)
        return self.approve(self.owner_address, bg)

    def create_token(self, price=0):
        return Token("", self.get_address(), price)


class BadERC20Token(IERC20Token):
    name = "BadERC20"
    protocol = token_protocol


class GenericToken(IERC20Token):
    name = "ERC20"
    protocol = token_protocol

    def __str__(self):
        return (
            f"GenericToken, name: {self._name().call()} address: {self.get_address()}"
        )

    def __repr__(self):
        return self.__str__()

    async def get_name(self) -> str:
        return await self._call_async(self._name())

    async def create_token(self, price=0):
        try:
            name = await self.get_name()
        except Exception:
            name = ""
            logging.getLogger().warning(
                f"Token: {self.get_address()} dosn't have a name."
            )
        return Token(name, self.get_address(), price)

    def _name(self):
        return self.contract.functions.name()


class Weth(GenericToken):
    name = "Weth"
    protocol = token_protocol

    def deposit(self, amount):
        transaction = self.contract.functions.deposit()
        return self._transact_status(transaction, BigNumber(amount).value)

    def withdraw(self, amount):
        transaction = self.contract.functions.withdraw(BigNumber(amount).value)
        return self._transact_status(transaction)



"""Integration test for Coingecko API."""

import pytest

from Arbie.Services import Coingecko

pytestmark = pytest.mark.asyncio


class TestCoingecko(object):
    async def test_btc_coin(self):
        response = await Coingecko()._coin_ticker("01coin")  # noqa: WPS437
        assert response is None

    async def test_eth_coin(self):
        address = await Coingecko()._coin_ticker("velo-token")  # noqa: WPS437
        assert address == "0x98ad9B32dD10f8D8486927D846D4Df8BAf39Abe2".lower()

    async def test_get_ids(self):
        ids = await Coingecko().ids()  # noqa: WPS437
        assert len(ids) > 6000  # noqa: WPS432
